#!/usr/bin/env python
"""
TODO
1. decrease memory usage
2. multi-fasta fastq file, ex. 454
3. split reads into small chuncks?

SHRiMP wrapper

Inputs:
1. reference seq
2. reads

Outputs:
1. table of 8 columns:
         chrom   ref_loc     read_id     read_loc    ref_nuc     read_nuc    quality     coverage
2. SHRiMP output

Parameters:
    -s    Spaced Seed                             (default: 111111011111)
    -n    Seed Matches per Window                 (default: 2)
    -t    Seed Hit Taboo Length                   (default: 4)
    -9    Seed Generation Taboo Length            (default: 0)
    -w    Seed Window Length                      (default: 115.00%)
    -o    Maximum Hits per Read                   (default: 100)
    -r    Maximum Read Length                     (default: 1000)
    -d    Kmer Std. Deviation Limit               (default: -1 [None])

    -m    S-W Match Value                         (default: 100)
    -i    S-W Mismatch Value                      (default: -150)
    -g    S-W Gap Open Penalty (Reference)        (default: -400)
    -q    S-W Gap Open Penalty (Query)            (default: -400)
    -e    S-W Gap Extend Penalty (Reference)      (default: -70)
    -f    S-W Gap Extend Penalty (Query)          (default: -70)
    -h    S-W Hit Threshold                       (default: 68.00%)

Command:
%rmapper -s spaced_seed -n seed_matches_per_window -t seed_hit_taboo_length -9 seed_generation_taboo_length -w seed_window_length -o max_hits_per_read -r max_read_length -d kmer -m sw_match_value -i sw_mismatch_value -g sw_gap_open_ref -q sw_gap_open_query -e sw_gap_ext_ref -f sw_gap_ext_query -h sw_hit_threshold <query> <target> > <output> 2> <log>

SHRiMP output:
>7:2:1147:982/1 chr3    +   36586562    36586595    2   35  36  2900    3G16G13
>7:2:1147:982/1 chr3    +   95338194    95338225    4   35  36  2700    9T7C14
>7:2:587:93/1   chr3    +   14913541    14913577    1   35  36  2960    19--16
"""
from __future__ import print_function

import os
import os.path
import re
import sys
import tempfile

assert sys.version_info[:2] >= (2, 6)


def stop_err(msg):
    sys.stderr.write("%s\n" % msg)
    sys.exit()


def reverse_complement(s):
    complement_dna = {
        "A": "T",
        "T": "A",
        "C": "G",
        "G": "C",
        "a": "t",
        "t": "a",
        "c": "g",
        "g": "c",
        "N": "N",
        "n": "n",
        ".": ".",
        "-": "-",
    }
    reversed_s = []
    for i in s:
        reversed_s.append(complement_dna[i])
    reversed_s.reverse()
    return "".join(reversed_s)


def generate_sub_table(result_file, ref_file, score_files, table_outfile, hit_per_read, insertion_size):
    invalid_editstring_char = 0
    all_score_file = score_files.split(",")

    if len(all_score_file) != hit_per_read:
        stop_err("One or more query files is missing. Please check your dataset.")

    temp_table_name = tempfile.NamedTemporaryFile().name
    temp_table = open(temp_table_name, "w")

    outfile = open(table_outfile, "w")

    # reference seq: not a single fasta seq
    refseq = {}
    chrom_cov = {}
    seq = ""
    title = None

    for line in open(ref_file):
        line = line.rstrip()
        if not line or line.startswith("#"):
            continue

        if line.startswith(">"):
            if seq:
                if title in refseq:
                    pass
                else:
                    refseq[title] = seq
                    chrom_cov[title] = {}
                seq = ""
            title = line[1:]
        else:
            seq += line
    if seq:
        if title not in refseq:
            refseq[title] = seq
            chrom_cov[title] = {}

    # find hits: one end and/or the other
    hits = {}
    for line in open(result_file):
        line = line.rstrip()
        if not line or line.startswith("#"):
            continue

        # FORMAT: readname contigname strand contigstart contigend readstart readend readlength score editstring
        fields = line.split("\t")
        readname = fields[0][1:]
        chrom = fields[1]
        strand = fields[2]
        chrom_start = int(fields[3]) - 1
        chrom_end = int(fields[4])
        read_start = fields[5]
        score = fields[8]
        editstring = fields[9]

        if hit_per_read == 1:
            endindex = "1"
        else:
            readname, endindex = readname.split("/")

        if readname in hits:
            if endindex in hits[readname]:
                hits[readname][endindex].append([strand, editstring, chrom_start, chrom_end, read_start, chrom])
            else:
                hits[readname][endindex] = [[strand, editstring, chrom_start, chrom_end, read_start, chrom]]
        else:
            hits[readname] = {}
            hits[readname][endindex] = [[strand, editstring, chrom_start, chrom_end, read_start, chrom]]

    # find score: one end and the other end
    hits_score = {}
    readname = ""
    score = ""
    for num_score_file in range(len(all_score_file)):
        score_file = all_score_file[num_score_file]
        for line in open(score_file):
            line = line.rstrip()
            if not line or line.startswith("#"):
                continue

            if line.startswith(">"):
                if score:
                    if readname in hits:
                        if len(hits[readname]) == hit_per_read:
                            if readname in hits_score:
                                if endindex in hits_score[readname]:
                                    pass
                                else:
                                    hits_score[readname][endindex] = score
                            else:
                                hits_score[readname] = {}
                                hits_score[readname][endindex] = score
                    score = ""
                if hit_per_read == 1:
                    readname = line[1:]
                    endindex = "1"
                else:
                    readname, endindex = line[1:].split("/")
            else:
                score = line

        if score:  # the last one
            if readname in hits:
                if len(hits[readname]) == hit_per_read:
                    if readname in hits_score:
                        if endindex in hits_score[readname]:
                            pass
                        else:
                            hits_score[readname][endindex] = score
                    else:
                        hits_score[readname] = {}
                        hits_score[readname][endindex] = score

    # call to all mappings
    for readkey in hits.keys():
        if len(hits[readkey]) != hit_per_read:
            continue

        matches = []
        match_count = 0

        if hit_per_read == 1:
            if len(hits[readkey]["1"]) == 1:
                matches = [hits[readkey]["1"]]
                match_count = 1
        else:
            end1_data = hits[readkey]["1"]
            end2_data = hits[readkey]["2"]

            for end1_hit in end1_data:
                crin_strand = {"+": False, "-": False}
                crin_insertSize = {"+": False, "-": False}

                crin_strand[end1_hit[0]] = True
                crin_insertSize[end1_hit[0]] = int(end1_hit[2])

                for end2_hit in end2_data:
                    crin_strand[end2_hit[0]] = True
                    crin_insertSize[end2_hit[0]] = int(end2_hit[2])

                    if end1_hit[-1] != end2_hit[-1]:
                        continue

                    if crin_strand["+"] and crin_strand["-"]:
                        if (crin_insertSize["-"] - crin_insertSize["+"]) <= insertion_size:
                            matches.append([end1_hit, end2_hit])
                            match_count += 1

        if match_count == 1:
            for x, end_data in enumerate(matches[0]):
                end_strand, end_editstring, end_chr_start, end_chr_end, end_read_start, end_chrom = end_data
                end_read_start = int(end_read_start) - 1

                if end_strand == "-":
                    refsegment = reverse_complement(refseq[end_chrom][end_chr_start:end_chr_end])
                else:
                    refsegment = refseq[end_chrom][end_chr_start:end_chr_end]

                match_len = 0
                editindex = 0
                gap_read = 0

                while editindex < len(end_editstring):
                    editchr = end_editstring[editindex]
                    chrA = ""
                    chrB = ""

                    if editchr.isdigit():
                        editcode = ""

                        while editchr.isdigit() and editindex < len(end_editstring):
                            editcode += editchr
                            editindex += 1
                            if editindex < len(end_editstring):
                                editchr = end_editstring[editindex]

                        for baseIndex in range(int(editcode)):
                            chrA += refsegment[match_len + baseIndex]
                            chrB = chrA

                        match_len += int(editcode)

                    elif editchr == "x":
                        # crossover: inserted between the appropriate two bases
                        # Two sequencing errors: 4x15x6 (25 matches with 2 crossovers)
                        # Treated as errors in the reads; Do nothing.
                        editindex += 1

                    elif editchr.isalpha():
                        editcode = editchr
                        editindex += 1
                        chrA = refsegment[match_len]
                        chrB = editcode
                        match_len += len(editcode)

                    elif editchr == "-":
                        editcode = editchr
                        editindex += 1
                        chrA = refsegment[match_len]
                        chrB = editcode
                        match_len += len(editcode)
                        gap_read += 1

                    elif editchr == "(":
                        editcode = ""

                        while editchr != ")" and editindex < len(end_editstring):
                            if editindex < len(end_editstring):
                                editchr = end_editstring[editindex]
                            editcode += editchr
                            editindex += 1

                        editcode = editcode[1:-1]
                        chrA = "-" * len(editcode)
                        chrB = editcode

                    else:
                        invalid_editstring_char += 1

                    if end_strand == "-":
                        chrA = reverse_complement(chrA)
                        chrB = reverse_complement(chrB)

                    pos_line = ""
                    rev_line = ""

                    for mappingIndex in range(len(chrA)):
                        # reference
                        chrAx = chrA[mappingIndex]
                        # read
                        chrBx = chrB[mappingIndex]

                        if chrAx and chrBx and chrBx.upper() != "N":
                            if end_strand == "+":
                                chrom_loc = end_chr_start + match_len - len(chrA) + mappingIndex
                                read_loc = end_read_start + match_len - len(chrA) + mappingIndex - gap_read

                                if chrAx == "-":
                                    chrom_loc -= 1

                                if chrBx == "-":
                                    scoreBx = "-1"
                                else:
                                    scoreBx = hits_score[readkey][str(x + 1)].split()[read_loc]

                                # 1-based on chrom_loc and read_loc
                                pos_line = (
                                    pos_line
                                    + "\t".join(
                                        (
                                            end_chrom,
                                            str(chrom_loc + 1),
                                            readkey + "/" + str(x + 1),
                                            str(read_loc + 1),
                                            chrAx,
                                            chrBx,
                                            scoreBx,
                                        )
                                    )
                                    + "\n"
                                )

                            else:
                                chrom_loc = end_chr_end - match_len + mappingIndex
                                read_loc = end_read_start + match_len - 1 - mappingIndex - gap_read

                                if chrAx == "-":
                                    chrom_loc -= 1

                                if chrBx == "-":
                                    scoreBx = "-1"
                                else:
                                    scoreBx = hits_score[readkey][str(x + 1)].split()[read_loc]

                                # 1-based on chrom_loc and read_loc
                                rev_line = (
                                    "\t".join(
                                        (
                                            end_chrom,
                                            str(chrom_loc + 1),
                                            readkey + "/" + str(x + 1),
                                            str(read_loc + 1),
                                            chrAx,
                                            chrBx,
                                            scoreBx,
                                        )
                                    )
                                    + "\n"
                                    + rev_line
                                )

                            if end_chrom in chrom_cov:
                                if chrom_loc in chrom_cov[end_chrom]:
                                    chrom_cov[end_chrom][chrom_loc] += 1
                                else:
                                    chrom_cov[end_chrom][chrom_loc] = 1

                            else:
                                chrom_cov[end_chrom] = {}
                                chrom_cov[end_chrom][chrom_loc] = 1

                    if pos_line:
                        temp_table.write("%s\n" % (pos_line.rstrip("\r\n")))
                    if rev_line:
                        temp_table.write("%s\n" % (rev_line.rstrip("\r\n")))

    temp_table.close()

    # chrom-wide coverage
    for line in open(temp_table_name):
        line = line.rstrip()
        if not line or line.startswith("#"):
            continue

        fields = line.split()
        chrom = fields[0]
        eachBp = int(fields[1])
        readname = fields[2]

        if hit_per_read == 1:
            fields[2] = readname.split("/")[0]

        if eachBp in chrom_cov[chrom]:
            outfile.write("%s\t%d\n" % ("\t".join(fields), chrom_cov[chrom][eachBp]))
        else:
            outfile.write("%s\t%d\n" % ("\t".join(fields), 0))

    outfile.close()

    if os.path.exists(temp_table_name):
        os.remove(temp_table_name)

    if invalid_editstring_char:
        print("Skip ", invalid_editstring_char, " invalid characters in editstrings")

    return True


def convert_fastqsolexa_to_fasta_qual(infile_name, query_fasta, query_qual):
    outfile_seq = open(query_fasta, "w")
    outfile_score = open(query_qual, "w")

    seq_title_startswith = ""
    qual_title_startswith = ""

    default_coding_value = 64  # Solexa ascii-code
    fastq_block_lines = 0

    for i, line in enumerate(open(infile_name)):
        line = line.rstrip()
        if not line or line.startswith("#"):
            continue

        fastq_block_lines = (fastq_block_lines + 1) % 4
        line_startswith = line[0:1]

        if fastq_block_lines == 1:
            # first line is @title_of_seq
            if not seq_title_startswith:
                seq_title_startswith = line_startswith

            if line_startswith != seq_title_startswith:
                outfile_seq.close()
                outfile_score.close()
                stop_err("Invalid fastqsolexa format at line %d: %s." % (i + 1, line))

            read_title = line[1:]
            outfile_seq.write(">%s\n" % line[1:])

        elif fastq_block_lines == 2:
            # second line is nucleotides
            read_length = len(line)
            outfile_seq.write("%s\n" % line)

        elif fastq_block_lines == 3:
            # third line is +title_of_qualityscore ( might be skipped )
            if not qual_title_startswith:
                qual_title_startswith = line_startswith

            if line_startswith != qual_title_startswith:
                outfile_seq.close()
                outfile_score.close()
                stop_err("Invalid fastqsolexa format at line %d: %s." % (i + 1, line))

            quality_title = line[1:]
            if quality_title and read_title != quality_title:
                outfile_seq.close()
                outfile_score.close()
                stop_err(
                    'Invalid fastqsolexa format at line %d: sequence title "%s" differes from score title "%s".'
                    % (i + 1, read_title, quality_title)
                )

            if not quality_title:
                outfile_score.write(">%s\n" % read_title)
            else:
                outfile_score.write(">%s\n" % line[1:])

        else:
            # fourth line is quality scores
            qual = ""
            fastq_integer = True
            # peek: ascii or digits?
            val = line.split()[0]
            fastq_integer = True
            try:
                int(val)
            except ValueError:
                fastq_integer = False

            if fastq_integer:
                # digits
                qual = line
            else:
                # ascii
                quality_score_length = len(line)
                if quality_score_length == read_length + 1:
                    # first char is qual_score_startswith
                    qual_score_startswith = ord(line[0:1])
                    line = line[1:]
                elif quality_score_length == read_length:
                    qual_score_startswith = default_coding_value
                else:
                    stop_err(
                        "Invalid fastqsolexa format at line %d: the number of quality scores ( %d ) is not the same as bases ( %d )."
                        % (i + 1, quality_score_length, read_length)
                    )

                for char in line:
                    score = ord(char) - qual_score_startswith  # 64
                    qual = "%s%s " % (qual, str(score))

            outfile_score.write("%s\n" % qual)

    outfile_seq.close()
    outfile_score.close()

    return True


def __main__():
    # SHRiMP path
    shrimp = "rmapper-ls"

    # I/O
    input_target_file = sys.argv[1]  # fasta
    shrimp_outfile = sys.argv[2]  # shrimp output
    table_outfile = sys.argv[3]  # table output
    single_or_paired = sys.argv[4].split(",")

    insertion_size = 600

    if len(single_or_paired) == 1:  # single or paired
        type_of_reads = "single"
        hit_per_read = 1
        input_query = single_or_paired[0]
        query_fasta = tempfile.NamedTemporaryFile().name
        query_qual = tempfile.NamedTemporaryFile().name

    else:  # paired-end
        type_of_reads = "paired"
        hit_per_read = 2
        input_query_end1 = single_or_paired[0]
        input_query_end2 = single_or_paired[1]
        insertion_size = int(single_or_paired[2])
        query_fasta_end1 = tempfile.NamedTemporaryFile().name
        query_fasta_end2 = tempfile.NamedTemporaryFile().name
        query_qual_end1 = tempfile.NamedTemporaryFile().name
        query_qual_end2 = tempfile.NamedTemporaryFile().name

    # SHRiMP parameters: total = 15, default values
    spaced_seed = "111111011111"
    seed_matches_per_window = "2"
    seed_hit_taboo_length = "4"
    seed_generation_taboo_length = "0"
    seed_window_length = "115.0"
    max_hits_per_read = "100"
    max_read_length = "1000"
    kmer = "-1"
    sw_match_value = "100"
    sw_mismatch_value = "-150"
    sw_gap_open_ref = "-400"
    sw_gap_open_query = "-400"
    sw_gap_ext_ref = "-70"
    sw_gap_ext_query = "-70"
    sw_hit_threshold = "68.0"

    # TODO: put the threshold on each of these parameters
    if len(sys.argv) > 5:
        try:
            if sys.argv[5].isdigit():
                spaced_seed = sys.argv[5]
            else:
                stop_err("Error in assigning parameter: Spaced seed.")
        except Exception:
            stop_err("Spaced seed must be a combination of 1s and 0s.")

        seed_matches_per_window = sys.argv[6]
        seed_hit_taboo_length = sys.argv[7]
        seed_generation_taboo_length = sys.argv[8]
        seed_window_length = sys.argv[9]
        max_hits_per_read = sys.argv[10]
        max_read_length = sys.argv[11]
        kmer = sys.argv[12]
        sw_match_value = sys.argv[13]
        sw_mismatch_value = sys.argv[14]
        sw_gap_open_ref = sys.argv[15]
        sw_gap_open_query = sys.argv[16]
        sw_gap_ext_ref = sys.argv[17]
        sw_gap_ext_query = sys.argv[18]
        sw_hit_threshold = sys.argv[19]

    # temp file for shrimp log file
    shrimp_log = tempfile.NamedTemporaryFile().name

    # convert fastq to fasta and quality score files
    if type_of_reads == "single":
        convert_fastqsolexa_to_fasta_qual(input_query, query_fasta, query_qual)
    else:
        convert_fastqsolexa_to_fasta_qual(input_query_end1, query_fasta_end1, query_qual_end1)
        convert_fastqsolexa_to_fasta_qual(input_query_end2, query_fasta_end2, query_qual_end2)

    # SHRiMP command
    if type_of_reads == "single":
        command = " ".join(
            (
                shrimp,
                "-s",
                spaced_seed,
                "-n",
                seed_matches_per_window,
                "-t",
                seed_hit_taboo_length,
                "-9",
                seed_generation_taboo_length,
                "-w",
                seed_window_length,
                "-o",
                max_hits_per_read,
                "-r",
                max_read_length,
                "-d",
                kmer,
                "-m",
                sw_match_value,
                "-i",
                sw_mismatch_value,
                "-g",
                sw_gap_open_ref,
                "-q",
                sw_gap_open_query,
                "-e",
                sw_gap_ext_ref,
                "-f",
                sw_gap_ext_query,
                "-h",
                sw_hit_threshold,
                query_fasta,
                input_target_file,
                ">",
                shrimp_outfile,
                "2>",
                shrimp_log,
            )
        )

        try:
            os.system(command)
        except Exception as e:
            if os.path.exists(query_fasta):
                os.remove(query_fasta)
            if os.path.exists(query_qual):
                os.remove(query_qual)
            stop_err(str(e))

    else:  # paired
        command_end1 = " ".join(
            (
                shrimp,
                "-s",
                spaced_seed,
                "-n",
                seed_matches_per_window,
                "-t",
                seed_hit_taboo_length,
                "-9",
                seed_generation_taboo_length,
                "-w",
                seed_window_length,
                "-o",
                max_hits_per_read,
                "-r",
                max_read_length,
                "-d",
                kmer,
                "-m",
                sw_match_value,
                "-i",
                sw_mismatch_value,
                "-g",
                sw_gap_open_ref,
                "-q",
                sw_gap_open_query,
                "-e",
                sw_gap_ext_ref,
                "-f",
                sw_gap_ext_query,
                "-h",
                sw_hit_threshold,
                query_fasta_end1,
                input_target_file,
                ">",
                shrimp_outfile,
                "2>",
                shrimp_log,
            )
        )
        command_end2 = " ".join(
            (
                shrimp,
                "-s",
                spaced_seed,
                "-n",
                seed_matches_per_window,
                "-t",
                seed_hit_taboo_length,
                "-9",
                seed_generation_taboo_length,
                "-w",
                seed_window_length,
                "-o",
                max_hits_per_read,
                "-r",
                max_read_length,
                "-d",
                kmer,
                "-m",
                sw_match_value,
                "-i",
                sw_mismatch_value,
                "-g",
                sw_gap_open_ref,
                "-q",
                sw_gap_open_query,
                "-e",
                sw_gap_ext_ref,
                "-f",
                sw_gap_ext_query,
                "-h",
                sw_hit_threshold,
                query_fasta_end2,
                input_target_file,
                ">>",
                shrimp_outfile,
                "2>>",
                shrimp_log,
            )
        )

        try:
            os.system(command_end1)
            os.system(command_end2)
        except Exception as e:
            if os.path.exists(query_fasta_end1):
                os.remove(query_fasta_end1)
            if os.path.exists(query_fasta_end2):
                os.remove(query_fasta_end2)
            if os.path.exists(query_qual_end1):
                os.remove(query_qual_end1)
            if os.path.exists(query_qual_end2):
                os.remove(query_qual_end2)
            stop_err(str(e))

    # check SHRiMP output: count number of lines
    num_hits = 0
    if shrimp_outfile:
        for line in open(shrimp_outfile):
            line = line.rstrip("\r\n")
            if not line or line.startswith("#"):
                continue
            try:
                line.split()
                num_hits += 1
            except Exception as e:
                stop_err(str(e))

    if num_hits == 0:  # no hits generated
        err_msg = ""
        if shrimp_log:
            for line in open(shrimp_log):
                if line.startswith("error"):  # deal with memory error:
                    err_msg += line  # error: realloc failed: Cannot allocate memory
                if re.search("Reads Matched", line):  # deal with zero hits
                    if int(line[8:].split()[2]) == 0:
                        err_msg = "Zero hits found.\n"
        stop_err("SHRiMP Failed due to:\n" + err_msg)

    # convert to table
    if type_of_reads == "single":
        generate_sub_table(shrimp_outfile, input_target_file, query_qual, table_outfile, hit_per_read, insertion_size)
    else:
        generate_sub_table(
            shrimp_outfile,
            input_target_file,
            query_qual_end1 + "," + query_qual_end2,
            table_outfile,
            hit_per_read,
            insertion_size,
        )

    # remove temp. files
    if type_of_reads == "single":
        if os.path.exists(query_fasta):
            os.remove(query_fasta)
        if os.path.exists(query_qual):
            os.remove(query_qual)
    else:
        if os.path.exists(query_fasta_end1):
            os.remove(query_fasta_end1)
        if os.path.exists(query_fasta_end2):
            os.remove(query_fasta_end2)
        if os.path.exists(query_qual_end1):
            os.remove(query_qual_end1)
        if os.path.exists(query_qual_end2):
            os.remove(query_qual_end2)

    if os.path.exists(shrimp_log):
        os.remove(shrimp_log)


if __name__ == "__main__":
    __main__()



from typing import (
    NamedTuple,
    Optional,
)

from galaxy.model import (
    DatasetInstance,
    HistoryDatasetAssociation,
)
from galaxy.security.idencoding import IdEncodingHelper


class ParameterOption(NamedTuple):
    name: str
    value: str
    selected: bool = False
    dataset: Optional["DatasetInstance"] = None

    def serialize(self, security: "IdEncodingHelper"):
        if self.dataset:
            return (
                self.name,
                {
                    "src": "hda" if isinstance(self.dataset, HistoryDatasetAssociation) else "ldda",
                    "id": security.encode_id(self.dataset.id),
                },
                self.selected,
            )
        return (self.name, self.value, self.selected)



import logging
from typing import (
    Any,
    Dict,
    List,
    Optional,
    overload,
    TYPE_CHECKING,
    Union,
)
from zipfile import ZipFile

from sqlalchemy.orm import (
    joinedload,
    Query,
)
from typing_extensions import Literal

from galaxy import model
from galaxy.datatypes.registry import Registry
from galaxy.exceptions import (
    ItemAccessibilityException,
    MessageException,
    RequestParameterInvalidException,
)
from galaxy.managers.collections_util import validate_input_element_identifiers
from galaxy.managers.context import (
    ProvidesAppContext,
    ProvidesHistoryContext,
    ProvidesUserContext,
)
from galaxy.managers.hdas import (
    HDAManager,
    HistoryDatasetAssociationNoHistoryException,
)
from galaxy.managers.hdcas import write_dataset_collection
from galaxy.managers.histories import HistoryManager
from galaxy.managers.lddas import LDDAManager
from galaxy.model import DatasetCollection
from galaxy.model.dataset_collections import builder
from galaxy.model.dataset_collections.matching import MatchingCollections
from galaxy.model.dataset_collections.registry import DATASET_COLLECTION_TYPES_REGISTRY
from galaxy.model.dataset_collections.type_description import COLLECTION_TYPE_DESCRIPTION_FACTORY
from galaxy.model.mapping import GalaxyModelMapping
from galaxy.schema.schema import (
    DatasetCollectionInstanceType,
    HistoryContentSource,
)
from galaxy.schema.tasks import PrepareDatasetCollectionDownload
from galaxy.security.idencoding import IdEncodingHelper
from galaxy.short_term_storage import (
    ShortTermStorageMonitor,
    storage_context,
)
from galaxy.util import validation

if TYPE_CHECKING:
    from galaxy.model import (
        DatasetCollectionInstance,
        HistoryDatasetAssociation,
    )
    from galaxy.tool_util_models.tool_source import FieldDict

log = logging.getLogger(__name__)

ERROR_INVALID_ELEMENTS_SPECIFICATION = "Create called with invalid parameters, must specify element identifiers."
ERROR_NO_COLLECTION_TYPE = "Create called without specifying a collection type."

HDCAElementObjectType = Union[DatasetCollection, "HistoryDatasetAssociation"]


class DatasetCollectionManager:
    """
    Abstraction for interfacing with dataset collections instance - ideally abstracts
    out model and plugin details.
    """

    ELEMENTS_UNINITIALIZED = object()

    def __init__(
        self,
        model: GalaxyModelMapping,
        security: IdEncodingHelper,
        hda_manager: HDAManager,
        history_manager: HistoryManager,
        ldda_manager: LDDAManager,
        short_term_storage_monitor: ShortTermStorageMonitor,
    ):
        self.type_registry = DATASET_COLLECTION_TYPES_REGISTRY
        self.collection_type_descriptions = COLLECTION_TYPE_DESCRIPTION_FACTORY
        self.model = model
        self.security = security
        self.short_term_storage_monitor = short_term_storage_monitor
        self.hda_manager = hda_manager
        self.history_manager = history_manager
        self.ldda_manager = ldda_manager

    def precreate_dataset_collection_instance(
        self,
        trans: ProvidesHistoryContext,
        parent,
        name,
        structure,
        implicit_inputs=None,
        implicit_output_name=None,
        tags=None,
        completed_collection=None,
    ):
        # TODO: prebuild all required HIDs and send them in so no need to flush in between.
        dataset_collection = self.precreate_dataset_collection(
            structure,
            allow_uninitialized_element=implicit_output_name is not None,
            completed_collection=completed_collection,
            implicit_output_name=implicit_output_name,
        )
        instance = self._create_instance_for_collection(
            trans,
            parent,
            name,
            dataset_collection,
            implicit_inputs=implicit_inputs,
            implicit_output_name=implicit_output_name,
            flush=False,
            tags=tags,
        )
        return instance

    def precreate_dataset_collection(
        self, structure, allow_uninitialized_element=True, completed_collection=None, implicit_output_name=None
    ):
        has_structure = not structure.is_leaf and structure.children_known
        if not has_structure and allow_uninitialized_element:
            dataset_collection = model.DatasetCollectionElement.UNINITIALIZED_ELEMENT
        elif not has_structure:
            collection_type_description = structure.collection_type_description
            dataset_collection = DatasetCollection(populated=False)
            dataset_collection.collection_type = collection_type_description.collection_type
        else:
            collection_type_description = structure.collection_type_description
            dataset_collection = DatasetCollection(populated=False)
            dataset_collection.collection_type = collection_type_description.collection_type
            elements = []
            for index, (identifier, substructure) in enumerate(structure.children):
                # TODO: Open question - populate these now or later?
                element = None
                if completed_collection and implicit_output_name:
                    job = completed_collection[index]
                    if job:
                        it = (
                            jtiodca.dataset_collection
                            for jtiodca in job.output_dataset_collections
                            if jtiodca.name == implicit_output_name
                        )
                        element = next(it, None)
                if element is None:
                    if substructure.is_leaf:
                        element = model.DatasetCollectionElement.UNINITIALIZED_ELEMENT
                    else:
                        element = self.precreate_dataset_collection(
                            substructure, allow_uninitialized_element=allow_uninitialized_element
                        )

                element = model.DatasetCollectionElement(
                    collection=dataset_collection,
                    element=element,
                    element_identifier=identifier,
                    element_index=index,
                )
                elements.append(element)
            dataset_collection.element_count = len(elements)

        return dataset_collection

    def create(
        self,
        trans: ProvidesHistoryContext,
        parent,
        name,
        collection_type,
        element_identifiers=None,
        elements=None,
        implicit_collection_info=None,
        trusted_identifiers=None,
        hide_source_items: bool = False,
        tags=None,
        copy_elements: bool = False,
        history=None,
        set_hid: bool = True,
        flush=True,
        completed_job=None,
        output_name=None,
        fields: Optional[Union[str, List["FieldDict"]]] = None,
    ) -> "DatasetCollectionInstance":
        """
        PRECONDITION: security checks on ability to add to parent
        occurred during load.
        """
        # Trust embedded, newly created objects created by tool subsystem.
        if trusted_identifiers is None:
            trusted_identifiers = implicit_collection_info is not None

        if element_identifiers and not trusted_identifiers:
            validate_input_element_identifiers(element_identifiers)

        if completed_job and output_name:
            jtodca = next(a for a in completed_job.output_dataset_collection_instances if a.name == output_name)
            dataset_collection = jtodca.dataset_collection_instance.collection
        else:
            dataset_collection = self.create_dataset_collection(
                trans=trans,
                collection_type=collection_type,
                element_identifiers=element_identifiers,
                elements=elements,
                hide_source_items=hide_source_items,
                copy_elements=copy_elements,
                history=history,
                fields=fields,
            )

        implicit_inputs = []
        if implicit_collection_info:
            implicit_inputs = implicit_collection_info.get("implicit_inputs", [])

        implicit_output_name = None
        if implicit_collection_info:
            implicit_output_name = implicit_collection_info["implicit_output_name"]

        return self._create_instance_for_collection(
            trans,
            parent,
            name,
            dataset_collection,
            implicit_inputs=implicit_inputs,
            implicit_output_name=implicit_output_name,
            tags=tags,
            set_hid=set_hid,
            flush=flush,
        )

    def _create_instance_for_collection(
        self,
        trans: ProvidesHistoryContext,
        parent,
        name,
        dataset_collection,
        implicit_output_name=None,
        implicit_inputs=None,
        tags=None,
        set_hid: bool = True,
        flush: bool = True,
    ) -> "DatasetCollectionInstance":
        if isinstance(parent, model.History):
            dataset_collection_instance: Union[
                model.HistoryDatasetCollectionAssociation,
                model.LibraryDatasetCollectionAssociation,
            ] = model.HistoryDatasetCollectionAssociation(
                collection=dataset_collection,
                name=name,
            )
            assert isinstance(dataset_collection_instance, model.HistoryDatasetCollectionAssociation)
            if implicit_inputs:
                for input_name, input_collection in implicit_inputs:
                    if isinstance(input_collection, model.HistoryDatasetCollectionAssociation):
                        # Can also get dragged DatasetCollectionElement's here.
                        # We only use this for extracting workflows currently,
                        # but obviously it would be better to track that somehow.
                        dataset_collection_instance.add_implicit_input_collection(input_name, input_collection)

            if implicit_output_name:
                dataset_collection_instance.implicit_output_name = implicit_output_name

            log.debug("Created collection with %d elements", len(dataset_collection_instance.collection.elements))

            if set_hid:
                parent.add_dataset_collection(dataset_collection_instance)

        elif isinstance(parent, model.LibraryFolder):
            dataset_collection_instance = model.LibraryDatasetCollectionAssociation(
                collection=dataset_collection,
                folder=parent,
                name=name,
            )

        else:
            message = f"Internal logic error - create called with unknown parent type {type(parent)}"
            log.exception(message)
            raise MessageException(message)

        # Tags may be coming in as a dictionary of tag model objects if copying them from other
        # existing Galaxy objects or as a list of strings if the tags are coming from user supplied
        # values.
        if isinstance(tags, list):
            assert implicit_inputs is None, implicit_inputs
            tags = trans.tag_handler.add_tags_from_list(trans.user, dataset_collection_instance, tags, flush=False)
        else:
            tags = self._append_tags(dataset_collection_instance, implicit_inputs, tags)
        return self.__persist(dataset_collection_instance, flush=flush)

    def create_dataset_collection(
        self,
        trans: ProvidesHistoryContext,
        collection_type,
        element_identifiers=None,
        elements=None,
        hide_source_items: bool = False,
        copy_elements: bool = False,
        history=None,
        fields: Optional[Union[str, List["FieldDict"]]] = None,
    ) -> DatasetCollection:
        # Make sure at least one of these is None.
        assert element_identifiers is None or elements is None
        if element_identifiers is None and elements is None:
            raise RequestParameterInvalidException(ERROR_INVALID_ELEMENTS_SPECIFICATION)
        if not collection_type:
            raise RequestParameterInvalidException(ERROR_NO_COLLECTION_TYPE)

        collection_type_description = self.collection_type_descriptions.for_collection_type(
            collection_type, fields=fields
        )
        has_subcollections = collection_type_description.has_subcollections()

        # If we have elements, this is an internal request, don't need to load
        # objects from identifiers.
        if elements is None:
            elements = self._element_identifiers_to_elements(
                trans,
                collection_type_description=collection_type_description,
                element_identifiers=element_identifiers,
                hide_source_items=hide_source_items,
                copy_elements=copy_elements,
                history=history,
            )
            if history:
                history.add_pending_items()
        else:
            if has_subcollections:
                # Nested collection - recursively create collections as needed.
                self.__recursively_create_collections_for_elements(
                    trans, elements, hide_source_items, copy_elements=copy_elements, history=history
                )
        # else if elements is set, it better be an ordered dict!

        if elements is not self.ELEMENTS_UNINITIALIZED:
            type_plugin = collection_type_description.rank_type_plugin()
            dataset_collection = builder.build_collection(type_plugin, elements, fields=fields)
        else:
            # TODO: Pass fields here - need test case first.
            dataset_collection = DatasetCollection(populated=False)
        dataset_collection.collection_type = collection_type
        return dataset_collection

    def get_converters_for_collection(
        self, trans: ProvidesHistoryContext, id, datatypes_registry: Registry, instance_type="history"
    ):
        dataset_collection_instance = self.get_dataset_collection_instance(
            trans, id=id, instance_type=instance_type, check_ownership=True
        )
        dbkeys_and_extensions = dataset_collection_instance.dataset_dbkeys_and_extensions_summary
        suitable_converters = set()
        first_extension = True
        most_recent_datatype = None
        # TODO error checking
        for datatype in dbkeys_and_extensions[1]:
            new_converters = datatypes_registry.get_converters_by_datatype(datatype)
            set_of_new_converters = set()
            for tgt_type, tgt_val in new_converters.items():
                converter = (tgt_type, tgt_val)
                set_of_new_converters.add(converter)
            if first_extension is True:
                suitable_converters = set_of_new_converters
                most_recent_datatype = datatype
                first_extension = False
            else:
                suitable_converters = suitable_converters.intersection(set_of_new_converters)
                if suitable_converters:
                    most_recent_datatype = datatype
        suitable_tool_ids = []
        for tool in suitable_converters:
            tool_info = {
                "tool_id": tool[1].id,
                "name": tool[1].name,
                "target_type": tool[0],
                "original_type": most_recent_datatype,
            }
            suitable_tool_ids.append(tool_info)
        return suitable_tool_ids

    def _element_identifiers_to_elements(
        self,
        trans: ProvidesHistoryContext,
        collection_type_description,
        element_identifiers,
        hide_source_items: bool = False,
        copy_elements: bool = False,
        history=None,
    ) -> Dict[str, HDCAElementObjectType]:
        if collection_type_description.has_subcollections():
            # Nested collection - recursively create collections and update identifiers.
            self.__recursively_create_collections_for_identifiers(
                trans, element_identifiers, hide_source_items, copy_elements, history=history
            )
        new_collection = False
        for element_identifier in element_identifiers:
            if element_identifier.get("src") == "new_collection" and element_identifier.get("collection_type") == "":
                new_collection = True
                elements = self.__load_elements(
                    trans=trans,
                    element_identifiers=element_identifier["element_identifiers"],
                    hide_source_items=hide_source_items,
                    copy_elements=copy_elements,
                    history=history,
                )
        if not new_collection:
            elements = self.__load_elements(
                trans=trans,
                element_identifiers=element_identifiers,
                hide_source_items=hide_source_items,
                copy_elements=copy_elements,
                history=history,
            )
        return elements

    def _append_tags(self, dataset_collection_instance, implicit_inputs=None, tags=None):
        tags = tags or {}
        implicit_inputs = implicit_inputs or []
        for _, v in implicit_inputs:
            for tag in v.auto_propagated_tags:
                tags[tag.value] = tag
        for _, tag in tags.items():
            dataset_collection_instance.tags.append(tag.copy(cls=model.HistoryDatasetCollectionTagAssociation))

    def collection_builder_for(self, dataset_collection):
        return builder.BoundCollectionBuilder(dataset_collection)

    def delete(self, trans: ProvidesHistoryContext, instance_type, id, recursive=False, purge=False):
        dataset_collection_instance = self.get_dataset_collection_instance(
            trans, instance_type, id, check_ownership=True
        )
        dataset_collection_instance.deleted = True
        trans.sa_session.add(dataset_collection_instance)
        async_result = None
        if recursive:
            for dataset in dataset_collection_instance.collection.dataset_instances:
                try:
                    self.hda_manager.error_unless_owner(dataset, user=trans.get_user(), current_history=trans.history)
                except HistoryDatasetAssociationNoHistoryException:
                    log.info(
                        f"Cannot delete HistoryDatasetAssociation {dataset.id}, HistoryDatasetAssociation has no associated History, cannot verify owner"
                    )
                    continue
                if not dataset.deleted:
                    dataset.deleted = True

                if purge and not dataset.purged:
                    async_result = self.hda_manager.purge(dataset, user=trans.user)

        trans.sa_session.commit()
        return async_result

    def update(self, trans: ProvidesHistoryContext, instance_type, id, payload):
        dataset_collection_instance = self.get_dataset_collection_instance(
            trans, instance_type, id, check_ownership=True
        )
        if trans.user is None:
            anon_allowed_payload = {}
            if "deleted" in payload:
                anon_allowed_payload["deleted"] = payload["deleted"]
            if "visible" in payload:
                anon_allowed_payload["visible"] = payload["visible"]
            payload = self._validate_and_parse_update_payload(anon_allowed_payload)
        else:
            payload = self._validate_and_parse_update_payload(payload)
        changed = self._set_from_dict(trans, dataset_collection_instance, payload)
        return changed

    def copy(
        self,
        trans: ProvidesHistoryContext,
        parent: model.History,
        source: Literal[HistoryContentSource.hdca],
        encoded_source_id,
        copy_elements: bool = False,
        dataset_instance_attributes: Optional[Dict[str, Any]] = None,
    ):
        """
        PRECONDITION: security checks on ability to add to parent occurred
        during load.
        """
        assert source == HistoryContentSource.hdca  # for now
        source_hdca = self.__get_history_collection_instance(trans, encoded_source_id)
        element_destination = parent if copy_elements else None
        new_hdca = source_hdca.copy(
            flush=False,
            element_destination=element_destination,
            dataset_instance_attributes=dataset_instance_attributes,
        )
        new_hdca.copy_tags_from(target_user=trans.get_user(), source=source_hdca)
        if not copy_elements:
            parent.add_dataset_collection(new_hdca)
        trans.sa_session.commit()
        return new_hdca

    def _set_from_dict(self, trans: ProvidesUserContext, dataset_collection_instance, new_data):
        # send what we can down into the model
        changed = dataset_collection_instance.set_from_dict(new_data)

        # the rest (often involving the trans) - do here
        if "annotation" in new_data.keys() and trans.get_user():
            dataset_collection_instance.add_item_annotation(
                trans.sa_session, trans.get_user(), dataset_collection_instance, new_data["annotation"]
            )
            changed["annotation"] = new_data["annotation"]

        # the api promises a list of changed fields, but tags are not marked as changed to avoid the
        # flush, so we must handle changed tag responses manually
        new_tags = None
        if "tags" in new_data.keys():
            # set_tags_from_list will flush on its own, no need to add to 'changed' here and incur a second flush.
            new_tags = trans.tag_handler.set_tags_from_list(
                trans.user,
                dataset_collection_instance,
                new_data["tags"],
            )

        if changed.keys():
            trans.sa_session.commit()

        # set client tag field response after the flush
        if new_tags is not None:
            changed["tags"] = dataset_collection_instance.make_tag_string_list()

        return changed

    def _validate_and_parse_update_payload(self, payload):
        validated_payload = {}
        for key, val in payload.items():
            if val is None:
                continue
            if key in ("name"):
                val = validation.validate_and_sanitize_basestring(key, val)
                validated_payload[key] = val
            if key in ("deleted", "visible"):
                validated_payload[key] = validation.validate_boolean(key, val)
            elif key == "tags":
                validated_payload[key] = validation.validate_and_sanitize_basestring_list(key, val)
        return validated_payload

    def history_dataset_collections(self, history, query):
        collections = history.active_dataset_collections
        collections = list(filter(query.direct_match, collections))
        return collections

    def __persist(
        self,
        dataset_collection_instance: "DatasetCollectionInstance",
        flush: bool = True,
    ) -> "DatasetCollectionInstance":
        session = self.model.session
        session.add(dataset_collection_instance)
        if flush:
            session.commit()
        return dataset_collection_instance

    def __recursively_create_collections_for_identifiers(
        self, trans, element_identifiers, hide_source_items: bool, copy_elements: bool, history=None
    ):
        for element_identifier in element_identifiers:
            try:
                if element_identifier.get("src") != "new_collection":
                    # not a new collection, keep moving...
                    continue
            except KeyError:
                # Not a dictionary, just an id of an HDA - move along.
                continue

            # element identifier is a dict with src new_collection...
            collection_type = element_identifier.get("collection_type")
            collection = self.create_dataset_collection(
                trans=trans,
                collection_type=collection_type,
                element_identifiers=element_identifier["element_identifiers"],
                hide_source_items=hide_source_items,
                copy_elements=copy_elements,
                history=history,
            )
            element_identifier["__object__"] = collection

        return element_identifiers

    def __recursively_create_collections_for_elements(
        self, trans, elements, hide_source_items: bool, copy_elements: bool, history=None
    ) -> None:
        if elements is self.ELEMENTS_UNINITIALIZED:
            return

        new_elements = {}
        for key, element in elements.items():
            if isinstance(element, DatasetCollection):
                continue

            if element.get("src") != "new_collection":
                continue

            # element is a dict with src new_collection and
            # and dict of named elements
            collection_type = element.get("collection_type")
            sub_elements = element["elements"]
            collection = self.create_dataset_collection(
                trans=trans,
                collection_type=collection_type,
                elements=sub_elements,
                hide_source_items=hide_source_items,
                copy_elements=copy_elements,
                history=history,
            )
            new_elements[key] = collection
        elements.update(new_elements)

    def __load_elements(
        self, trans, element_identifiers, hide_source_items: bool = False, copy_elements: bool = False, history=None
    ) -> Dict[str, HDCAElementObjectType]:
        elements: Dict[str, HDCAElementObjectType] = {}
        for element_identifier in element_identifiers:
            elements[element_identifier["name"]] = self.__load_element(
                trans,
                element_identifier=element_identifier,
                hide_source_items=hide_source_items,
                copy_elements=copy_elements,
                history=history,
            )
        return elements

    def __load_element(
        self, trans, element_identifier, hide_source_items: bool, copy_elements: bool, history=None
    ) -> HDCAElementObjectType:
        # if not isinstance( element_identifier, dict ):
        #    # Is allowing this to just be the id of an hda too clever? Somewhat
        #    # consistent with other API methods though.
        #    element_identifier = dict( src='hda', id=str( element_identifier ) )

        # Previously created collection already found in request, just pass
        # through as is.
        if "__object__" in element_identifier:
            the_object = element_identifier["__object__"]
            if the_object is not None and the_object.id:
                session = self.model.session
                if the_object not in session:
                    the_object = session.get(type(the_object), the_object.id)
            return the_object

        # dataset_identifier is dict {src=hda|ldda|hdca|new_collection, id=<encoded_id>}
        try:
            src_type = element_identifier.get("src", "hda")
        except AttributeError:
            raise MessageException(f"Dataset collection element definition ({element_identifier}) not dictionary-like.")
        element_id = element_identifier.get("id")
        if not src_type or not element_id:
            message_template = "Problem decoding element identifier %s - must contain a 'src' and a 'id'."
            message = message_template % element_identifier
            raise RequestParameterInvalidException(message)

        tag_str = ""
        if tags := element_identifier.pop("tags", None):
            tag_str = ",".join(str(_) for _ in tags)
        if src_type == "hda":
            hda = self.hda_manager.get_accessible(element_id, trans.user)
            if copy_elements:
                element = self.hda_manager.copy(hda, history=history or trans.history, hide_copy=True, flush=False)
            else:
                element = hda
            if hide_source_items and self.hda_manager.get_owned(
                hda.id, user=trans.user, current_history=history or trans.history
            ):
                hda.visible = False
            trans.tag_handler.apply_item_tags(user=trans.user, item=element, tags_str=tag_str, flush=False)
            return element
        elif src_type == "ldda":
            element2 = self.ldda_manager.get(trans, element_id, check_accessible=True)
            element3 = element2.to_history_dataset_association(
                history or trans.history, add_to_history=True, visible=not hide_source_items
            )
            trans.tag_handler.apply_item_tags(user=trans.user, item=element3, tags_str=tag_str, flush=False)
            return element3
        elif src_type == "hdca":
            # TODO: Option to copy? Force copy? Copy or allow if not owned?
            return self.__get_history_collection_instance(trans, element_id).collection
        # TODO: ldca.
        raise RequestParameterInvalidException(f"Unknown src_type parameter supplied '{src_type}'.")

    def match_collections(self, collections_to_match):
        """
        May seem odd to place it here, but planning to grow sophistication and
        get plugin types involved so it will likely make sense in the future.
        """
        return MatchingCollections.for_collections(collections_to_match, self.collection_type_descriptions)

    @overload
    def get_dataset_collection_instance(
        self, trans: ProvidesHistoryContext, instance_type: Literal["history"], id, **kwds: Any
    ) -> model.HistoryDatasetCollectionAssociation: ...

    @overload
    def get_dataset_collection_instance(
        self, trans: ProvidesHistoryContext, instance_type: Literal["library"], id, **kwds: Any
    ) -> model.LibraryDatasetCollectionAssociation: ...

    def get_dataset_collection_instance(
        self, trans: ProvidesHistoryContext, instance_type: DatasetCollectionInstanceType, id, **kwds: Any
    ) -> "DatasetCollectionInstance":
        """ """
        if instance_type == "history":
            return self.__get_history_collection_instance(trans, id, **kwds)
        elif instance_type == "library":
            return self.__get_library_collection_instance(trans, id, **kwds)
        raise NotImplementedError()

    def get_dataset_collection(self, trans, encoded_id):
        collection_id = int(trans.app.security.decode_id(encoded_id))
        collection = trans.sa_session.get(DatasetCollection, collection_id)
        return collection

    def apply_rules(self, hdca, rule_set, handle_dataset):
        hdca_collection = hdca.collection
        collection_type = hdca_collection.collection_type
        elements = hdca_collection.elements
        collection_type_description = self.collection_type_descriptions.for_collection_type(collection_type)
        initial_data, initial_sources = self.__init_rule_data(elements, collection_type_description)
        data, sources = rule_set.apply(initial_data, initial_sources)

        collection_type = rule_set.collection_type
        collection_type_description = self.collection_type_descriptions.for_collection_type(collection_type)
        elements = self._build_elements_from_rule_data(
            collection_type_description, rule_set, data, sources, handle_dataset
        )
        return elements

    def _build_elements_from_rule_data(self, collection_type_description, rule_set, data, sources, handle_dataset):
        identifier_columns = rule_set.identifier_columns
        mapping_as_dict = rule_set.mapping_as_dict
        elements: Dict[str, Any] = {}
        for data_index, row_data in enumerate(data):
            # For each row, find place in depth for this element.
            collection_type_at_depth = collection_type_description
            elements_at_depth = elements

            for i, identifier_column in enumerate(identifier_columns):
                identifier = row_data[identifier_column]

                if i + 1 == len(identifier_columns):
                    # At correct final position in nested structure for this dataset.
                    if collection_type_at_depth.collection_type == "paired":
                        if identifier.lower() in ["f", "1", "r1", "forward"]:
                            identifier = "forward"
                        elif identifier.lower() in ["r", "2", "r2", "reverse"]:
                            identifier = "reverse"
                        else:
                            raise Exception(
                                "Unknown indicator of paired status encountered - only values of F, R, 1, 2, R1, R2, forward, or reverse are allowed."
                            )

                    tags = []
                    if "group_tags" in mapping_as_dict:
                        columns = mapping_as_dict["group_tags"]["columns"]
                        for tag_column in columns:
                            tag = row_data[tag_column]
                            tags.append(f"group:{tag}")

                    if "tags" in mapping_as_dict:
                        columns = mapping_as_dict["tags"]["columns"]
                        for tag_column in columns:
                            tag = row_data[tag_column]
                            tags.append(tag)

                    effective_dataset = handle_dataset(sources[data_index]["dataset"], tags)
                    elements_at_depth[identifier] = effective_dataset
                    # log.info("Handling dataset [%s] with sources [%s], need to add tags [%s]" % (effective_dataset, sources, tags))
                else:
                    collection_type_at_depth = collection_type_at_depth.child_collection_type_description()
                    found = False
                    if identifier in elements_at_depth:
                        elements_at_depth = elements_at_depth[identifier]["elements"]
                        found = True

                    if not found:
                        # Create a new collection whose elements are defined in the next loop
                        sub_collection: Dict[str, Any] = {}
                        sub_collection["src"] = "new_collection"
                        sub_collection["collection_type"] = collection_type_at_depth.collection_type
                        sub_collection["elements"] = {}
                        # Update elements with new collection
                        elements_at_depth[identifier] = sub_collection
                        # Subsequent loop fills elements of newly created collection
                        elements_at_depth = sub_collection["elements"]

        return elements

    def __init_rule_data(self, elements, collection_type_description, parent_identifiers=None, parent_indices=None):
        parent_identifiers = parent_identifiers or []
        parent_indices = parent_indices or []
        data: List[List[str]] = []
        sources: List[Dict[str, str]] = []
        for i, element in enumerate(elements):
            indices = parent_indices.copy()
            indices.append(i)

            element_object = element.element_object
            identifiers = parent_identifiers + [element.element_identifier]
            if not element.is_collection:
                data.append([])
                source = {
                    "identifiers": identifiers,
                    "dataset": element_object,
                    "tags": element_object.make_tag_string_list(),
                    "indices": indices,
                }
                sources.append(source)
            else:
                child_collection_type_description = collection_type_description.child_collection_type_description()
                element_data, element_sources = self.__init_rule_data(
                    element_object.elements,
                    child_collection_type_description,
                    identifiers,
                    parent_indices=indices,
                )
                data.extend(element_data)
                sources.extend(element_sources)

        return data, sources

    def __get_history_collection_instance(
        self, trans: ProvidesHistoryContext, id, check_ownership=False, check_accessible=True
    ) -> model.HistoryDatasetCollectionAssociation:
        instance_id = trans.app.security.decode_id(id) if isinstance(id, str) else id
        collection_instance = trans.sa_session.get(model.HistoryDatasetCollectionAssociation, instance_id)
        if not collection_instance:
            raise RequestParameterInvalidException("History dataset collection association not found")
        # TODO: that sure looks like a bug, we can't check ownership using the history of the object we're checking ownership for ...
        history = getattr(trans, "history", collection_instance.history)
        if check_ownership:
            self.history_manager.error_unless_owner(collection_instance.history, trans.user, current_history=history)
        if check_accessible:
            self.history_manager.error_unless_accessible(
                collection_instance.history, trans.user, current_history=history
            )
        return collection_instance

    def __get_library_collection_instance(
        self, trans: ProvidesHistoryContext, id, check_ownership=False, check_accessible=True
    ) -> model.LibraryDatasetCollectionAssociation:
        if check_ownership:
            raise NotImplementedError(
                "Functionality (getting library dataset collection with ownership check) unimplemented."
            )
        instance_id = int(trans.security.decode_id(id))
        collection_instance = trans.sa_session.get(model.LibraryDatasetCollectionAssociation, instance_id)
        if not collection_instance:
            raise RequestParameterInvalidException("Library dataset collection association not found")
        if check_accessible:
            if not trans.app.security_agent.can_access_library_item(
                trans.get_current_user_roles(), collection_instance, trans.user
            ):
                raise ItemAccessibilityException(
                    "LibraryDatasetCollectionAssociation is not accessible to the current user", type="error"
                )
        return collection_instance

    def get_collection_contents(self, trans: ProvidesAppContext, parent_id, limit=None, offset=None):
        """Find first level of collection contents by containing collection parent_id"""
        contents_qry = self._get_collection_contents_qry(parent_id, limit=limit, offset=offset)
        contents = contents_qry.with_session(trans.sa_session()).all()
        return contents

    def _get_collection_contents_qry(self, parent_id, limit=None, offset=None):
        """Build query to find first level of collection contents by containing collection parent_id"""
        DCE = model.DatasetCollectionElement
        qry = Query(DCE).filter(DCE.dataset_collection_id == parent_id)  # type:ignore[var-annotated]
        qry = qry.order_by(DCE.element_index)
        qry = qry.options(
            joinedload(model.DatasetCollectionElement.child_collection), joinedload(model.DatasetCollectionElement.hda)
        )
        if limit is not None:
            qry = qry.limit(int(limit))
        if offset is not None:
            qry = qry.offset(int(offset))
        return qry

    def write_dataset_collection(self, request: PrepareDatasetCollectionDownload):
        short_term_storage_monitor = self.short_term_storage_monitor
        instance_id = request.history_dataset_collection_association_id
        with storage_context(request.short_term_storage_request_id, short_term_storage_monitor) as target:
            collection_instance = self.model.context.get(model.HistoryDatasetCollectionAssociation, instance_id)
            with ZipFile(target.path, "w") as zip_f:
                write_dataset_collection(collection_instance, zip_f)



#!/usr/bin/env python
"""
Adapted from bx/scripts/axt_to_fasta.py
"""
from __future__ import print_function

import sys

import bx.align.axt


def usage(s=None):
    message = """
axt_to_fasta species1 species2 < axt_file > fasta_file
"""
    if s is None:
        sys.exit(message)
    else:
        sys.exit("%s\n%s" % (s, message))


def main():
    # check the command line
    species1 = sys.argv[1]
    species2 = sys.argv[2]

    # convert the alignment blocks

    reader = bx.align.axt.Reader(sys.stdin, support_ids=True, species1=species1, species2=species2)

    for a in reader:
        if "id" in a.attributes:
            id = a.attributes["id"]
        else:
            id = None
        print_component_as_fasta(a.components[0], id)
        print_component_as_fasta(a.components[1], id)
        print()


# TODO: this should be moved to a bx.align.fasta module
def print_component_as_fasta(c, id=None):
    header = ">%s_%s_%s" % (c.src, c.start, c.start + c.size)
    if id is not None:
        header += " " + id
    print(header)
    print(c.text)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Guruprasad Ananda
# Refactored 2011 to use numpy instead of rpy, Kanwei Li
"""
This tool provides the SQL "group by" functionality.
Arguments:
    1 output file name
    2 input file name
    3 grouping column
    4 ignore case (1/0)
    5 ascii to delete (comma separated list)
    6... op,col,do_round,default
"""
from __future__ import print_function

import random
import subprocess
import sys
import tempfile
from itertools import groupby

import numpy


def float_wdefault(s, d, c):
    """
    convert list of strings s into list of floats
    non convertable entries are replaced by d if d is not None (otherwise error)
    """
    for i in range(len(s)):
        try:
            s[i] = float(s[i])
        except ValueError:
            if d is not None:
                s[i] = d
            else:
                stop_err("non float value '%s' found in colum %d" % (s[i], c))
    return s


def stop_err(msg):
    sys.stderr.write(msg)
    sys.exit(1)


def mode(data):
    counts = {}
    for x in data:
        counts[x] = counts.get(x, 0) + 1
    maxcount = max(counts.values())
    modelist = []
    for x in counts:
        if counts[x] == maxcount:
            modelist.append(str(x))
    return ",".join(modelist)


def main():
    inputfile = sys.argv[2]
    ignorecase = int(sys.argv[4])
    ops = []
    cols = []
    round_val = []
    default_val = []

    # remove comment lines
    if sys.argv[5] != "None":
        asciitodelete = sys.argv[5]
        if asciitodelete:
            newinputfile = "input_cleaned.tsv"
            with open(inputfile) as oldfile, open(newinputfile, "w") as newfile:
                asciitodelete = {chr(int(_)) for _ in asciitodelete.split(",")}
                for line in oldfile:
                    if line[0] not in asciitodelete:
                        newfile.write(line)
            inputfile = newinputfile

    # get operations and options in separate arrays
    for var in sys.argv[6:]:
        op, col, do_round, default = var.split(",")
        ops.append(op)
        cols.append(col)
        round_val.append(do_round)
        default_val.append(float(default) if default != "" else None)

    # At this point, ops, cols and rounds will look something like this:
    # ops:  ['mean', 'min', 'c']
    # cols: ['1', '3', '4']
    # round_val: ['no', 'yes' 'no']
    # default_val: [0, 1, None]

    try:
        group_col = int(sys.argv[3]) - 1
    except Exception:
        stop_err("Group column not specified.")

    # sort file into a temporary file
    tmpfile = tempfile.NamedTemporaryFile(mode="r")
    try:
        """
        The -k option for the Posix sort command is as follows:
        -k, --key=POS1[,POS2]
        start a key at POS1, end it at POS2 (origin 1)
        In other words, column positions start at 1 rather than 0, so
        we need to add 1 to group_col.
        if POS2 is not specified, the newer versions of sort will consider the entire line for sorting. To prevent this, we set POS2=POS1.
        """
        group_col_str = str(group_col + 1)
        command_line = ["sort", "-t", "\t", "-k%s,%s" % (group_col_str, group_col_str), "-o", tmpfile.name, inputfile]
        if ignorecase == 1:
            command_line.append("-f")
    except Exception as exc:
        stop_err("Initialization error -> %s" % str(exc))

    try:
        subprocess.check_output(command_line, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        stop_err("Sorting input dataset resulted in error: %s: %s" % (e.returncode, e.output.decode()))

    def is_new_item(line):
        try:
            item = line.rstrip("\r\n").split("\t")[group_col]
        except IndexError:
            stop_err("The following line didn't have %s columns: %s" % (group_col + 1, line))
        if ignorecase == 1:
            return item.lower()
        return item

    with open(sys.argv[1], "w") as fout:
        for key, line_list in groupby(tmpfile, key=is_new_item):
            op_vals = [[] for _ in ops]
            out_str = key

            for line in line_list:
                fields = line.split("\t")
                for i, col in enumerate(cols):
                    col = int(col) - 1  # cXX from galaxy is 1-based
                    try:
                        val = fields[col].strip()
                        op_vals[i].append(val)
                    except IndexError:
                        sys.stderr.write(
                            'Could not access the value for column %s on line: "%s". Make sure file is tab-delimited.\n'
                            % (col + 1, line)
                        )
                        sys.exit(1)

            # Generate string for each op for this group
            for i, op in enumerate(ops):
                data = op_vals[i]
                rval = ""
                if op == "mode":
                    rval = mode(data)
                elif op == "length":
                    rval = len(data)
                elif op == "random":
                    rval = random.choice(data)
                elif op in ["cat", "cat_uniq"]:
                    if op == "cat_uniq":
                        data = numpy.unique(data)
                    rval = ",".join(data)
                elif op == "unique":
                    rval = len(numpy.unique(data))
                else:
                    # some kind of numpy fn
                    try:
                        data = float_wdefault(data, default_val[i], col + 1)
                    except ValueError:
                        sys.stderr.write("Operation %s expected number values but got %s instead.\n" % (op, data))
                        sys.exit(1)
                    rval = getattr(numpy, op)(data)
                    if round_val[i] == "yes":
                        rval = int(round(rval))
                    else:
                        rval = "%g" % rval
                out_str += "\t%s" % rval

            fout.write(out_str + "\n")

    tmpfile.close()

    # Generate a useful info message.
    msg = "--Group by c%d: " % (group_col + 1)
    for i, op in enumerate(ops):
        if op == "cat":
            op = "concat"
        elif op == "cat_uniq":
            op = "concat_distinct"
        elif op == "length":
            op = "count"
        elif op == "unique":
            op = "count_distinct"
        elif op == "random":
            op = "randomly_pick"

        msg += op + "[c" + cols[i] + "] "

    print(msg)


if __name__ == "__main__":
    main()



import re

from .newickparser import Newick_Parser

MAX_READLINES = 200000


class Nexus_Parser(Newick_Parser):
    def __init__(self):
        super().__init__()

    def parseFile(self, filePath):
        """passes a file and extracts its Nexus content."""
        return self.parseNexus(filePath)

    def parseNexus(self, filename):
        """Nexus data is stored in blocks between a line starting with begin and another line starting with end;
        Commends inside square brackets are to be ignored,
        For more information: http://wiki.christophchamp.com/index.php/NEXUS_file_format
        Nexus can store multiple trees
        """

        with open(filename) as nex_file:
            nexlines = nex_file.readlines()

        rowCount = 0
        inTreeBlock = False  # sentinel to check if we are in a tree block
        intranslateBlock = (
            False  # sentinel to check if we are in the translate region of the tree. Stores synonyms of the labellings
        )
        self.inCommentBlock = False
        self.nameMapping = None  # stores mapping representation used in nexus format
        treeNames = []

        for line in nexlines:
            line = line.replace(";\n", "")
            lline = line.lower()

            if rowCount > MAX_READLINES or (not nex_file):
                break
            rowCount += 1
            # We are only interested in the tree block.
            if "begin" in lline and "tree" in lline and not inTreeBlock:
                inTreeBlock = True
                continue
            if inTreeBlock and "end" in lline[:3]:
                inTreeBlock, currPhyloTree = False, None
                continue

            if inTreeBlock:
                if "title" in lline:  # Adding title to the tree
                    continue

                if "translate" in lline:
                    intranslateBlock = True
                    self.nameMapping = {}
                    continue

                if intranslateBlock:
                    mappingLine = self.splitLinebyWhitespaces(line)
                    key, value = mappingLine[1], mappingLine[2].replace(",", "").replace(
                        "'", ""
                    )  # replacing illegal json characters
                    self.nameMapping[key] = value

                # Extracting newick Trees
                if "tree" in lline:
                    intranslateBlock = False

                    treeLineCols = self.splitLinebyWhitespaces(line)
                    treeName, newick = treeLineCols[2], treeLineCols[-1]

                    if newick == "":  # Empty lines can be found in tree blocks
                        continue

                    currPhyloTree = self._parseNewickToJson(newick, treeName, nameMap=self.nameMapping)

                    self.phyloTrees.append(currPhyloTree)
                    treeIndex = len(self.phyloTrees) - 1
                    treeNames.append((treeName, treeIndex))  # appending name of tree, and its index
                    continue

        return self.phyloTrees, treeNames

    def splitLinebyWhitespaces(self, line):
        """replace tabs and write spaces to a single write space, so we can properly split it."""
        return re.split(r"\s+", line)

    def checkComments(self, line):
        """Check to see if the line/lines is a comment."""
        if not self.inCommentBlock:
            if "[" in line:
                if "]" not in line:
                    self.inCommentBlock = True
                else:
                    return "Nextline"  # need to move on to the nextline after getting out of comment
        else:
            if "]" in line:
                if line.rfind("[") > line.rfind("]"):
                    pass  # a comment block is closed but another is open.
                else:
                    self.inCommentBlock = False
                    return "Nextline"  # need to move on to the nextline after getting out of comment
        return ""



from galaxy import util
from galaxy.util.tool_shed.common_util import get_tool_shed_url_from_tool_shed_registry


def get_updated_changeset_revisions_from_tool_shed(app, tool_shed_url, name, owner, changeset_revision):
    """
    Get all appropriate newer changeset revisions for the repository defined by
    the received tool_shed_url / name / owner combination.
    """
    tool_shed_url = get_tool_shed_url_from_tool_shed_registry(app, tool_shed_url)
    if tool_shed_url is None or name is None or owner is None or changeset_revision is None:
        message = "Unable to get updated changeset revisions from the Tool Shed because one or more of the following "
        message += f"required parameters is None: tool_shed_url: {tool_shed_url}, name: {name}, owner: {owner}, changeset_revision: {changeset_revision}"
        raise Exception(message)
    params = dict(name=name, owner=owner, changeset_revision=changeset_revision)
    pathspec = ["repository", "updated_changeset_revisions"]
    text = util.url_get(
        tool_shed_url, auth=app.tool_shed_registry.url_auth(tool_shed_url), pathspec=pathspec, params=params
    )
    return text



from typing import Union

from sqlalchemy import (
    and_,
    select,
)
from sqlalchemy.exc import (
    MultipleResultsFound,
    NoResultFound,
)

from galaxy.exceptions import (
    InconsistentDatabase,
    InternalServerError,
    RequestParameterInvalidException,
)
from galaxy.managers.context import ProvidesUserContext
from galaxy.model import ChatExchange
from galaxy.util import unicodify


class ChatManager:
    """
    Business logic for chat exchanges.
    """

    def create(self, trans: ProvidesUserContext, job_id: int, message: str) -> ChatExchange:
        """
        Create a new chat exchange in the DB.  Currently these are *only* job-based chat exchanges, will need to generalize down the road.
        :param  job_id:      id of the job to associate the response with
        :type   job_id:      int
        :param  response:    the response to save in the DB
        :type   response:    str
        :returns:   the created ChatExchange object
        :rtype:     galaxy.model.ChatExchange
        :raises: InternalServerError
        """
        chat_exchange = ChatExchange(user=trans.user, job_id=job_id, message=message)
        trans.sa_session.add(chat_exchange)
        trans.sa_session.commit()
        return chat_exchange

    def get(self, trans: ProvidesUserContext, job_id: int) -> Union[ChatExchange, None]:
        """
        Returns the chat response from the DB based on the given job id.
        :param  job_id:      id of the job to load a response for from the DB
        :type   job_id:      int
        :returns:   the loaded ChatExchange object
        :rtype:     galaxy.model.ChatExchange
        :raises: InconsistentDatabase, InternalServerError
        """
        try:
            stmt = select(ChatExchange).where(
                and_(ChatExchange.job_id == job_id, ChatExchange.user_id == trans.user.id)
            )
            chat_response = trans.sa_session.execute(stmt).scalar_one()
        except MultipleResultsFound:
            # TODO: Unsure about this, isn't this more applicable when we're getting the response for response.id instead of response.
            raise InconsistentDatabase("Multiple chat responses found with the same job id.")
        except NoResultFound:
            # TODO: Would there be cases where we raise an exception here? Or, is there a better way to return None?
            # raise RequestParameterInvalidException("No accessible response found with the id provided.")
            return None
        except Exception as e:
            raise InternalServerError(f"Error loading from the database.{unicodify(e)}")
        return chat_response

    def set_feedback_for_job(self, trans: ProvidesUserContext, job_id: int, feedback: int) -> ChatExchange:
        """
        Set the feedback for a chat response.
        :param  message_id:      id of the job to associate the feedback with
        :type   message_id:      int
        :param  feedback:    the feedback to save in the DB (0 or 1)
        :type   feedback:    int
        :returns:   the updated ChatExchange object
        :rtype:     galaxy.model.ChatExchange
        :raises: RequestParameterInvalidException
        """

        # TODO: Set feedback for specific messages as we allow multiple messages per exchange, not this method targeting job.

        # Validate the feedback; it should be 0 or 1
        if feedback not in [0, 1]:
            raise RequestParameterInvalidException("Feedback should be 0 or 1.")

        chat_exchange = self.get(trans, job_id)

        if not chat_exchange:
            raise RequestParameterInvalidException("No accessible response found with the id provided.")

        # There is only one message in an exchange currently, so we can set the feedback on the first message
        chat_exchange.messages[0].feedback = feedback

        trans.sa_session.commit()

        return chat_exchange



"""Helpers meant to assist tool execution.

Lower-level things that prevent interwoven dependencies between tool code,
tool execution code, and tool action code.
"""

import logging
from typing import Collection

log = logging.getLogger(__name__)


class ToolExecutionCache:
    """An object meant to cache calculation caused by repeatedly evaluting
    the same tool by the same user with slightly different parameters.
    """

    def __init__(self, trans):
        self.trans = trans
        self.current_user_roles = trans.get_current_user_roles()
        self.chrom_info = {}
        self.cached_collection_elements = {}

    def get_chrom_info(self, tool_id, input_dbkey):
        genome_builds = self.trans.app.genome_builds
        custom_build_hack_get_len_from_fasta_conversion = tool_id != "CONVERTER_fasta_to_len"
        if custom_build_hack_get_len_from_fasta_conversion and input_dbkey in self.chrom_info:
            return self.chrom_info[input_dbkey]

        chrom_info_pair = genome_builds.get_chrom_info(
            input_dbkey,
            trans=self.trans,
            custom_build_hack_get_len_from_fasta_conversion=custom_build_hack_get_len_from_fasta_conversion,
        )
        if custom_build_hack_get_len_from_fasta_conversion:
            self.chrom_info[input_dbkey] = chrom_info_pair

        return chrom_info_pair


def filter_output(tool, output, incoming):
    for filter in output.filters:
        try:
            if not eval(filter.text.strip(), globals(), incoming):
                return True  # do not create this dataset
        except Exception as e:
            log.debug(f"Tool {tool.id} output {output.name}: dataset output filter ({filter.text}) failed: {e}")
    return False


def on_text_for_names(input_names: Collection[str]) -> str:
    # input_names may contain duplicates... this is because the first value in
    # multiple input dataset parameters will appear twice once as param_name
    # and once as param_name1.
    unique_names = []
    for name in input_names:
        if name not in unique_names:
            unique_names.append(name)
    input_names = unique_names

    # Build name for output datasets based on tool name and input names
    if len(input_names) == 0:
        on_text = ""
    elif len(input_names) == 1:
        on_text = input_names[0]
    elif len(input_names) == 2:
        on_text = "{} and {}".format(*input_names)
    elif len(input_names) == 3:
        on_text = "{}, {}, and {}".format(*input_names)
    else:
        on_text = "{}, {}, and others".format(*input_names[:2])
    return on_text



"""
Manager and Serializer for libraries.
"""

import logging
from typing import (
    Dict,
    Optional,
    Set,
    Tuple,
)

from sqlalchemy.exc import (
    MultipleResultsFound,
    NoResultFound,
)
from sqlalchemy.orm import Query

from galaxy import exceptions
from galaxy.managers.folders import FolderManager
from galaxy.model import (
    Library,
    LibraryFolder,
    Role,
)
from galaxy.model.db.library import (
    get_libraries_by_name,
    get_libraries_for_admins,
    get_libraries_for_nonadmins,
    get_library_ids,
    get_library_permissions_by_role,
)
from galaxy.model.db.role import get_private_role_user_emails_dict
from galaxy.util import (
    pretty_print_time_interval,
    unicodify,
)

log = logging.getLogger(__name__)


# =============================================================================
class LibraryManager:
    """
    Interface/service object for interacting with libraries.
    """

    def get(self, trans, decoded_library_id: int, check_accessible: bool = True) -> Library:
        """
        Get the library from the DB.

        :param  decoded_library_id:       decoded library id
        :type   decoded_library_id:       int
        :param  check_accessible:         flag whether to check that user can access item
        :type   check_accessible:         bool

        :returns:   the requested library
        :rtype:     galaxy.model.Library
        """
        try:
            library = trans.sa_session.get(Library, decoded_library_id)
        except MultipleResultsFound:
            raise exceptions.InconsistentDatabase("Multiple libraries found with the same id.")
        except NoResultFound:
            raise exceptions.RequestParameterInvalidException("No library found with the id provided.")
        except Exception as e:
            raise exceptions.InternalServerError(f"Error loading from the database.{unicodify(e)}")
        library = self.secure(trans, library, check_accessible)
        return library

    def create(self, trans, name: str, description: Optional[str] = "", synopsis: Optional[str] = "") -> Library:
        """
        Create a new library.
        """
        if not trans.user_is_admin:
            raise exceptions.ItemAccessibilityException("Only administrators can create libraries.")
        else:
            library = Library(name=name, description=description, synopsis=synopsis)
            root_folder = LibraryFolder(name=name, description="")
            library.root_folder = root_folder
            trans.sa_session.add_all((library, root_folder))
            trans.sa_session.commit()
            return library

    def update(
        self,
        trans,
        library: Library,
        name: Optional[str] = None,
        description: Optional[str] = None,
        synopsis: Optional[str] = None,
    ) -> Library:
        """
        Update the given library
        """
        changed = False
        if not trans.user_is_admin:
            current_user_roles = trans.get_current_user_roles()
            library_modify_roles = self.get_modify_roles(trans, library)
            user_can_modify = any(role in library_modify_roles for role in current_user_roles)
            if not user_can_modify:
                raise exceptions.ItemAccessibilityException("You don't have permission update libraries.")
        if library.deleted:
            raise exceptions.RequestParameterInvalidException("You cannot modify a deleted library. Undelete it first.")
        if name is not None:
            library.name = name
            changed = True
            #  When library is renamed the root folder has to be renamed too.
            folder_manager = FolderManager()
            folder_manager.update(trans, library.root_folder, name=name)
        if description is not None:
            library.description = description
            changed = True
        if synopsis is not None:
            library.synopsis = synopsis
            changed = True
        if changed:
            trans.sa_session.add(library)
            trans.sa_session.commit()
        return library

    def delete(self, trans, library: Library, undelete: Optional[bool] = False) -> Library:
        """
        Mark given library deleted/undeleted based on the flag.
        """
        if not trans.user_is_admin:
            raise exceptions.ItemAccessibilityException("Only administrators can delete and undelete libraries.")
        if undelete:
            library.deleted = False
        else:
            library.deleted = True
        trans.sa_session.add(library)
        trans.sa_session.commit()
        return library

    def list(self, trans, deleted: Optional[bool] = False) -> Tuple[Query, Dict[str, Set]]:
        """
        Return a list of libraries from the DB.

        :param  deleted: if True, show only ``deleted`` libraries, if False show only ``non-deleted``
        :type   deleted: boolean (optional)

        :returns: iterable that will emit all accessible libraries
        :rtype:   sqlalchemy ScalarResult
        :returns: dict of 3 sets with available actions for user's accessible
                  libraries and a set of ids of all public libraries. These are
                  used for limiting the number of queries when dictifying the
                  libraries later on.
        :rtype:   dict
        """
        is_admin = trans.user_is_admin
        library_access_action = trans.app.security_agent.permitted_actions.LIBRARY_ACCESS.action
        restricted_library_ids = set(get_library_ids(trans.sa_session, library_access_action))
        prefetched_ids = {"restricted_library_ids": restricted_library_ids}

        if is_admin:
            libraries = get_libraries_for_admins(trans.sa_session, deleted=deleted)
        else:
            #  Nonadmins can't see deleted libraries
            if deleted:
                raise exceptions.AdminRequiredException()
            current_user_role_ids = [role.id for role in trans.get_current_user_roles()]
            library_add_action = trans.app.security_agent.permitted_actions.LIBRARY_ADD.action
            library_modify_action = trans.app.security_agent.permitted_actions.LIBRARY_MODIFY.action
            library_manage_action = trans.app.security_agent.permitted_actions.LIBRARY_MANAGE.action
            accessible_restricted_library_ids = set()
            allowed_library_add_ids = set()
            allowed_library_modify_ids = set()
            allowed_library_manage_ids = set()
            for action in get_library_permissions_by_role(trans.sa_session, current_user_role_ids):
                if action.action == library_access_action:
                    accessible_restricted_library_ids.add(action.library_id)
                if action.action == library_add_action:
                    allowed_library_add_ids.add(action.library_id)
                if action.action == library_modify_action:
                    allowed_library_modify_ids.add(action.library_id)
                if action.action == library_manage_action:
                    allowed_library_manage_ids.add(action.library_id)
            prefetched_ids["allowed_library_add_ids"] = allowed_library_add_ids
            prefetched_ids["allowed_library_modify_ids"] = allowed_library_modify_ids
            prefetched_ids["allowed_library_manage_ids"] = allowed_library_manage_ids
            libraries = get_libraries_for_nonadmins(
                trans.sa_session, restricted_library_ids, accessible_restricted_library_ids
            )

        return libraries, prefetched_ids

    def secure(self, trans, library: Library, check_accessible: bool = True) -> Library:
        """
        Check if library is accessible to user.

        :param  library:                 library
        :type   library:                 galaxy.model.Library
        :param  check_accessible:        flag whether to check that user can access library
        :type   check_accessible:        bool

        :returns:   the original library
        :rtype:     galaxy.model.Library
        """
        # all libraries are accessible to an admin
        if trans.user_is_admin:
            return library
        if check_accessible:
            library = self.check_accessible(trans, library)
        return library

    def check_accessible(self, trans, library: Library) -> Library:
        """
        Check whether the library is accessible to current user.
        """
        if not trans.app.security_agent.can_access_library(trans.get_current_user_roles(), library):
            raise exceptions.ObjectNotFound("Library with the id provided was not found.")
        elif library.deleted:
            raise exceptions.ObjectNotFound("Library with the id provided is deleted.")
        else:
            return library

    def get_library_dict(self, trans, library: Library, prefetched_ids: Optional[Dict[str, Set]] = None) -> dict:
        """
        Return library data in the form of a dictionary.

        :param  library:         library
        :type   library:         galaxy.model.Library
        :param  prefetched_ids:  dict of 3 sets with available actions for user's accessible
                                 libraries and a set of ids of all public libraries. These are
                                 used for limiting the number of queries when dictifying a
                                 set of libraries.
        :type   prefetched_ids:  dict

        :returns:   dict with data about the library
        :rtype:     dictionary
        """
        restricted_library_ids = prefetched_ids.get("restricted_library_ids", None) if prefetched_ids else None
        allowed_library_add_ids = prefetched_ids.get("allowed_library_add_ids", None) if prefetched_ids else None
        allowed_library_modify_ids = prefetched_ids.get("allowed_library_modify_ids", None) if prefetched_ids else None
        allowed_library_manage_ids = prefetched_ids.get("allowed_library_manage_ids", None) if prefetched_ids else None
        library_dict = library.to_dict(view="element")
        library_dict["public"] = False if (restricted_library_ids and library.id in restricted_library_ids) else True
        library_dict["create_time_pretty"] = pretty_print_time_interval(library.create_time, precise=True)
        if not trans.user_is_admin:
            if prefetched_ids:
                library_dict["can_user_add"] = (
                    True if (allowed_library_add_ids and library.id in allowed_library_add_ids) else False
                )
                library_dict["can_user_modify"] = (
                    True if (allowed_library_modify_ids and library.id in allowed_library_modify_ids) else False
                )
                library_dict["can_user_manage"] = (
                    True if (allowed_library_manage_ids and library.id in allowed_library_manage_ids) else False
                )
            else:
                current_user_roles = trans.get_current_user_roles()
                library_dict["can_user_add"] = trans.app.security_agent.can_add_library_item(
                    current_user_roles, library
                )
                library_dict["can_user_modify"] = trans.app.security_agent.can_modify_library_item(
                    current_user_roles, library
                )
                library_dict["can_user_manage"] = trans.app.security_agent.can_manage_library_item(
                    current_user_roles, library
                )
        else:
            library_dict["can_user_add"] = True
            library_dict["can_user_modify"] = True
            library_dict["can_user_manage"] = True
        return library_dict

    def get_current_roles(self, trans, library: Library) -> dict:
        """
        Load all permissions currently related to the given library.

        :param  library:      the model object
        :type   library:      galaxy.model.Library

        :rtype:     dictionary
        :returns:   dict of current roles for all available permission types
        """
        private_role_emails = get_private_role_user_emails_dict(trans.sa_session)
        access_roles = self.get_access_roles(trans, library)
        modify_roles = self.get_modify_roles(trans, library)
        manage_roles = self.get_manage_roles(trans, library)
        add_roles = self.get_add_roles(trans, library)

        def make_tuples(roles: Set):
            tuples = []
            for role in roles:
                # use role name for non-private roles, and user.email from private rules
                displayed_name = private_role_emails.get(role.id, role.name)
                role_tuple = (displayed_name, trans.security.encode_id(role.id))
                tuples.append(role_tuple)
            return tuples

        return dict(
            access_library_role_list=make_tuples(access_roles),
            modify_library_role_list=make_tuples(modify_roles),
            manage_library_role_list=make_tuples(manage_roles),
            add_library_item_role_list=make_tuples(add_roles),
        )

    def get_access_roles(self, trans, library: Library) -> Set[Role]:
        """
        Load access roles for all library permissions
        """
        return set(library.get_access_roles(trans.app.security_agent))

    def get_modify_roles(self, trans, library: Library) -> Set[Role]:
        """
        Load modify roles for all library permissions
        """
        return set(
            trans.app.security_agent.get_roles_for_action(
                library, trans.app.security_agent.permitted_actions.LIBRARY_MODIFY
            )
        )

    def get_manage_roles(self, trans, library: Library) -> Set[Role]:
        """
        Load manage roles for all library permissions
        """
        return set(
            trans.app.security_agent.get_roles_for_action(
                library, trans.app.security_agent.permitted_actions.LIBRARY_MANAGE
            )
        )

    def get_add_roles(self, trans, library: Library) -> Set[Role]:
        """
        Load add roles for all library permissions
        """
        return set(
            trans.app.security_agent.get_roles_for_action(
                library, trans.app.security_agent.permitted_actions.LIBRARY_ADD
            )
        )

    def make_public(self, trans, library: Library) -> bool:
        """
        Makes the given library public (removes all access roles)
        """
        trans.app.security_agent.make_library_public(library)
        return self.is_public(trans, library)

    def is_public(self, trans, library: Library) -> bool:
        """
        Return true if lib is public.
        """
        return trans.app.security_agent.library_is_public(library)


def get_containing_library_from_library_dataset(trans, library_dataset) -> Optional[Library]:
    """Given a library_dataset, get the containing library"""
    folder = library_dataset.folder
    while folder.parent:
        folder = folder.parent
    # We have folder set to the library's root folder, which has the same name as the library
    for library in get_libraries_by_name(trans.sa_session, folder.name):
        # Just to double-check
        if library.root_folder == folder:
            return library
    return None



import json
import logging
from urllib.parse import (
    urlsplit,
    urlunsplit,
)

from sqlalchemy import (
    Column,
    create_engine,
    delete,
    insert,
    Integer,
    MetaData,
    select,
    String,
    Table,
    Text,
)

from galaxy import exceptions
from galaxy.model import (
    InteractiveToolEntryPoint,
    Job,
)
from galaxy.security.idencoding import IdAsLowercaseAlphanumEncodingHelper

log = logging.getLogger(__name__)

gxitproxy = Table(
    "gxitproxy",
    MetaData(),
    Column("key", String(16), primary_key=True),
    Column("key_type", Text(), primary_key=True),
    Column("token", String(32)),
    Column("host", Text()),
    Column("port", Integer()),
    Column("info", Text()),
)


class InteractiveToolPropagatorSQLAlchemy:
    """
    Propagator for InteractiveToolManager implemented using SQLAlchemy.
    """

    def __init__(self, database_url, encode_id):
        """
        Constructor that sets up the propagator using a SQLAlchemy database URL.

        :param database_url: SQLAlchemy database URL, read more on the SQLAlchemy documentation
                             https://docs.sqlalchemy.org/en/20/core/engines.html#database-urls.
        :param encode_id: A helper class that can encode ids as lowercase alphanumeric strings and vice versa.
        """
        self._engine = create_engine(database_url)
        self._encode_id = encode_id

    def save(self, key, key_type, token, host, port, info=None):
        """
        Write out a key, key_type, token, value store that is can be used for coordinating with external resources.
        """
        assert key, ValueError("A non-zero length key is required.")
        assert key_type, ValueError("A non-zero length key_type is required.")
        assert token, ValueError("A non-zero length token is required.")
        with self._engine.connect() as conn:
            # create database table if not exists
            gxitproxy.create(conn, checkfirst=True)

            # delete existing data with same key
            stmt_delete = delete(gxitproxy).where(
                gxitproxy.c.key == key,
                gxitproxy.c.key_type == key_type,
            )
            conn.execute(stmt_delete)

            # save data
            stmt_insert = insert(gxitproxy).values(
                key=key,
                key_type=key_type,
                token=token,
                host=host,
                port=port,
                info=info,
            )
            conn.execute(stmt_insert)

            conn.commit()

    def remove(self, **kwd):
        """
        Remove entry from a key, key_type, token, value store that is can be used for coordinating
        with external resources. Remove entries that match all provided key=values
        """
        assert kwd, ValueError("You must provide some values to key upon")
        with self._engine.connect() as conn:
            stmt = delete(gxitproxy).where(
                *(gxitproxy.c[key] == value for key, value in kwd.items()),
            )
            conn.execute(stmt)
            conn.commit()

    def save_entry_point(self, entry_point):
        """Convenience method to easily save an entry_point."""
        return self.save(
            self._encode_id(entry_point.id),
            entry_point.__class__.__name__.lower(),
            entry_point.token,
            entry_point.host,
            entry_point.port,
            json.dumps(
                {
                    "requires_path_in_url": entry_point.requires_path_in_url,
                    "requires_path_in_header_named": entry_point.requires_path_in_header_named,
                }
            ),
        )

    def remove_entry_point(self, entry_point):
        """Convenience method to easily remove an entry_point."""
        return self.remove(key=self._encode_id(entry_point.id), key_type=entry_point.__class__.__name__.lower())


class InteractiveToolManager:
    """
    Manager for dealing with InteractiveTools
    """

    def __init__(self, app):
        self.app = app
        self.model = app.model
        self.security = app.security
        self.sa_session = app.model.context
        self.job_manager = app.job_manager
        self.encoder = IdAsLowercaseAlphanumEncodingHelper(app.security)
        self.propagator = InteractiveToolPropagatorSQLAlchemy(
            app.config.interactivetoolsproxy_map or app.config.interactivetools_map,
            self.encoder.encode_id,
        )

    def create_entry_points(self, job, tool, entry_points=None, flush=True):
        entry_points = entry_points or tool.ports
        for entry in entry_points:
            ep = self.model.InteractiveToolEntryPoint(
                job=job,
                tool_port=entry["port"],
                entry_url=entry["url"],
                name=entry["name"],
                label=entry["label"],
                requires_domain=entry["requires_domain"],
                requires_path_in_url=entry["requires_path_in_url"],
                requires_path_in_header_named=entry["requires_path_in_header_named"],
            )
            self.sa_session.add(ep)
        if flush:
            self.sa_session.commit()

    def configure_entry_point(self, job, tool_port=None, host=None, port=None, protocol=None):
        return self.configure_entry_points(
            job, {tool_port: dict(tool_port=tool_port, host=host, port=port, protocol=protocol)}
        )

    def configure_entry_points(self, job, ports_dict):
        # There can be multiple entry points that reference the same tool port (could have different entry URLs)
        configured = []
        not_configured = []
        for ep in job.interactivetool_entry_points:
            port_dict = ports_dict.get(str(ep.tool_port), None)
            if port_dict is None:
                log.error("Did not find port to assign to InteractiveToolEntryPoint by tool port: %s.", ep.tool_port)
                not_configured.append(ep)
            else:
                ep.host = port_dict["host"]
                ep.port = port_dict["port"]
                ep.protocol = port_dict["protocol"]
                ep.configured = True
                self.sa_session.add(ep)
                self.save_entry_point(ep)
                configured.append(ep)
        if configured:
            self.sa_session.commit()
        return dict(not_configured=not_configured, configured=configured)

    def save_entry_point(self, entry_point):
        """
        Writeout a key, key_type, token, value store that is used validating access
        """
        self.propagator.save_entry_point(entry_point)

    def create_interactivetool(self, job, tool, entry_points):
        # create from initial job
        if job and tool:
            self.create_entry_points(job, tool, entry_points)
        else:
            log.warning(
                "Called InteractiveToolManager.create_interactivetool, but job (%s) or tool (%s) is None", job, tool
            )

    def get_nonterminal_for_user_by_trans(self, trans):
        if trans.user is None and trans.get_galaxy_session() is None:
            return []

        stmt = (
            select(InteractiveToolEntryPoint)
            .join(Job, InteractiveToolEntryPoint.job_id == Job.id)
            .where(Job.state.in_(Job.non_ready_states))
        )
        if trans.user:
            stmt = stmt.where(Job.user == trans.user)
        else:
            stmt = stmt.where(Job.session_id == trans.get_galaxy_session().id)
        return trans.sa_session.scalars(stmt)

    def can_access_job(self, trans, job):
        if job:
            if trans.user is None:
                galaxy_session = trans.get_galaxy_session()
                if galaxy_session is None or job.session_id != galaxy_session.id:
                    return False
            elif job.user != trans.user:
                return False
        else:
            return False
        return True

    def can_access_entry_point(self, trans, entry_point):
        if entry_point:
            return self.can_access_job(trans, entry_point.job)
        return False

    def can_access_entry_points(self, trans, entry_points):
        for ep in entry_points:
            if not self.can_access_entry_point(trans, ep):
                return False
        return True

    def stop(self, trans, entry_point):
        self.remove_entry_point(entry_point)
        job = entry_point.job
        if not job.finished:
            log.debug("Stopping Job: %s for InteractiveToolEntryPoint: %s", job, entry_point)
            job.mark_stopped(trans.app.config.track_jobs_in_database)
            # This self.job_manager.stop(job) does nothing without changing job.state, manually or e.g. with .mark_deleted()
            self.job_manager.stop(job)
            trans.sa_session.add(job)
            trans.sa_session.commit()

    def remove_entry_points(self, entry_points):
        if entry_points:
            for entry_point in entry_points:
                self.remove_entry_point(entry_point, flush=False)
            self.sa_session.commit()

    def remove_entry_point(self, entry_point, flush=True):
        entry_point.deleted = True
        self.sa_session.add(entry_point)
        if flush:
            self.sa_session.commit()
        self.propagator.remove_entry_point(entry_point)

    def target_if_active(self, trans, entry_point):
        if entry_point.active and not entry_point.deleted:
            use_it_proxy_host_cfg = (
                not self.app.config.interactivetools_upstream_proxy and self.app.config.interactivetools_proxy_host
            )

            url_parts = urlsplit(trans.request.host_url)
            url_host = self.app.config.interactivetools_proxy_host if use_it_proxy_host_cfg else trans.request.host
            url_path = url_parts.path

            if entry_point.requires_domain:
                url_host = f"{self.get_entry_point_subdomain(trans, entry_point)}.{url_host}"
                if entry_point.entry_url:
                    url_path = f"{url_path.rstrip('/')}/{entry_point.entry_url.lstrip('/')}"
            else:
                url_path = self.get_entry_point_path(trans, entry_point)
                if not use_it_proxy_host_cfg:
                    return url_path

            return urlunsplit((url_parts.scheme, url_host, url_path, "", ""))

    def _get_entry_point_url_elements(self, trans, entry_point):
        encoder = IdAsLowercaseAlphanumEncodingHelper(trans.security)
        ep_encoded_id = encoder.encode_id(entry_point.id)
        ep_class_id = entry_point.class_id
        ep_prefix = self.app.config.interactivetools_prefix
        ep_token = entry_point.token
        return ep_encoded_id, ep_class_id, ep_prefix, ep_token

    def get_entry_point_subdomain(self, trans, entry_point):
        ep_encoded_id, ep_class_id, ep_prefix, ep_token = self._get_entry_point_url_elements(trans, entry_point)
        return f"{ep_encoded_id}-{ep_token}.{ep_class_id}.{ep_prefix}"

    def get_entry_point_path(self, trans, entry_point):
        url_path = "/"
        if not entry_point.requires_domain:
            ep_encoded_id, ep_class_id, ep_prefix, ep_token = self._get_entry_point_url_elements(trans, entry_point)
            path_parts = [
                part.strip("/")
                for part in (
                    str(self.app.config.interactivetools_base_path),
                    ep_prefix,
                    ep_class_id,
                    ep_encoded_id,
                    ep_token,
                )
            ]
            url_path += "/".join(part for part in path_parts if part) + "/"
        if entry_point.entry_url:
            url_path += entry_point.entry_url.lstrip("/")
        return url_path

    def access_entry_point_target(self, trans, entry_point_id):
        entry_point = trans.sa_session.get(InteractiveToolEntryPoint, entry_point_id)
        if self.app.interactivetool_manager.can_access_entry_point(trans, entry_point):
            if entry_point.active:
                return self.target_if_active(trans, entry_point)
            elif entry_point.deleted:
                raise exceptions.MessageException("InteractiveTool has ended. You will have to start a new one.")
            else:
                raise exceptions.MessageException(
                    "InteractiveTool is not active. If you recently launched this tool it may not be ready yet, please wait a moment and refresh this page."
                )
        raise exceptions.ItemAccessibilityException("You do not have access to this InteractiveTool entry point.")



from abc import (
    ABCMeta,
    abstractmethod,
)
from uuid import UUID

from celery.result import AsyncResult

from galaxy.schema.tasks import (
    TaskResult,
    TaskState,
)


class AsyncTasksManager(metaclass=ABCMeta):
    @abstractmethod
    def is_ready(self, task_uuid: UUID) -> bool:
        """Return `True` if the task has executed.

        If the task is still running, pending, or is waiting for retry then `False` is returned.
        """

    def is_pending(self, task_uuid: UUID) -> bool:
        """Return `True` if the task is not done yet."""
        return not self.is_ready(task_uuid)

    @abstractmethod
    def is_successful(self, task_uuid: UUID) -> bool:
        """Return `True` if the task executed successfully."""

    @abstractmethod
    def has_failed(self, task_uuid: UUID) -> bool:
        """Return `True` if the task failed."""

    @abstractmethod
    def get_state(self, task_uuid: UUID) -> TaskState:
        """Returns the current state of the task as a string."""

    @abstractmethod
    def get_result(self, task_uuid: UUID) -> TaskResult:
        """Returns the final state and result message of the task."""


class CeleryAsyncTasksManager(AsyncTasksManager):
    """Thin wrapper around Celery tasks AsyncResult queries."""

    def is_ready(self, task_uuid: UUID) -> bool:
        """Return `True` if the task has executed.

        If the task is still running, pending, or is waiting for retry then `False` is returned.
        """
        return self._get_result(task_uuid).ready()

    def is_successful(self, task_uuid: UUID) -> bool:
        """Return `True` if the task executed successfully."""
        return self._get_result(task_uuid).successful()

    def has_failed(self, task_uuid: UUID) -> bool:
        """Return `True` if the task failed."""
        return self._get_result(task_uuid).failed()

    def get_state(self, task_uuid: UUID) -> TaskState:
        """Returns the tasks current state as a string."""
        result = self._get_result(task_uuid)
        state = TaskState(result.state)
        return state

    def get_result(self, task_uuid: UUID) -> TaskResult:
        async_result = self._get_result(task_uuid)
        result = TaskResult(state=TaskState(async_result.state), result=str(async_result.result or ""))
        return result

    def _get_result(self, task_uuid: UUID) -> AsyncResult:
        return AsyncResult(str(task_uuid))



import logging
from typing import (
    List,
    Optional,
)

from pydantic import BaseModel

from galaxy.datatypes.registry import Registry
from galaxy.structured_app import StructuredApp

log = logging.getLogger(__name__)


class Link(BaseModel):
    name: str


class DisplayApplication(BaseModel):
    id: str
    name: str
    version: str
    filename_: str
    links: List[Link]


class ReloadFeedback(BaseModel):
    message: str
    reloaded: List[Optional[str]]
    failed: List[Optional[str]]


class DisplayApplicationsManager:
    """Interface/service object for sharing logic between controllers."""

    def __init__(self, app: StructuredApp):
        self._app = app

    @property
    def datatypes_registry(self) -> Registry:
        return self._app.datatypes_registry

    def index(self) -> List[DisplayApplication]:
        """
        Returns the list of display applications.

        :returns:   list of available display applications
        :rtype:     list
        """
        rval = []
        for display_app in self.datatypes_registry.display_applications.values():
            rval.append(
                DisplayApplication(
                    id=display_app.id,
                    name=display_app.name,
                    version=display_app.version,
                    filename_=display_app._filename,
                    links=[Link(name=link.name) for link in display_app.links.values()],
                )
            )
        return rval

    def reload(self, ids: List[str]) -> ReloadFeedback:
        """
        Reloads the list of display applications.

        :param  ids:  list containing ids of display to be reloaded
        :type   ids:  list
        """
        self._app.queue_worker.send_control_task(
            "reload_display_application", noop_self=True, kwargs={"display_application_ids": ids}
        )
        reloaded, failed = self.datatypes_registry.reload_display_applications(ids)
        if not reloaded and failed:
            message = 'Unable to reload any of the {} requested display applications ("{}").'.format(
                len(failed),
                '", "'.join(failed),
            )
        elif failed:
            message = (
                'Reloaded {} display applications ("{}"), but failed to reload {} display applications ("{}").'.format(
                    len(reloaded), '", "'.join(reloaded), len(failed), '", "'.join(failed)
                )
            )
        elif not reloaded:
            message = "You need to request at least one display application to reload."
        else:
            message = 'Reloaded {} requested display applications ("{}").'.format(len(reloaded), '", "'.join(reloaded))
        return ReloadFeedback(message=message, reloaded=reloaded, failed=failed)



#!/usr/bin/env python
"""
Script that imports locally stored data as a new dataset for the user
Usage: import id outputfile
"""
from __future__ import print_function

import os
import sys

assert sys.version_info[:2] >= (2, 6)

BUFFER = 1048576

dataid = sys.argv[1]
out_name = sys.argv[2]


id2name = {
    "eryth": "ErythPreCRMmm3_cusTrk.txt",
    "cishg16": "ReglRegHBBhg16CusTrk.txt",
    "cishg17": "ReglRegHBBhg17CusTrk.txt",
    "exons": "ExonsKnownGenes_mm3.txt",
    "krhg16": "known_regulatory_hg16.bed",
    "krhg17": "known_regulatory_hg17.bed",
    "tARhg16mmc": "hg16.mouse.t_AR.cold.bed",
    "tARhg16mmm": "hg16.mouse.t_AR.medium.bed",
    "tARhg16mmh": "hg16.mouse.t_AR.hot.bed",
    "tARhg16rnc": "hg16.rat.t_AR.cold.bed",
    "tARhg16rnm": "hg16.rat.t_AR.medium.bed",
    "tARhg16rnh": "hg16.rat.t_AR.hot.bed",
    "phastConsHg16": "phastConsMost_hg16.bed",
    "omimhg16": "omimDisorders_hg16.tab",
    "omimhg17": "omimDisorders_hg17.tab",
}

fname = id2name.get(dataid, "")
if not fname:
    print("Importing invalid data %s" % dataid)
    sys.exit()
else:
    print("Imported %s" % fname)

# this path is hardcoded
inp_name = os.path.join("database", "import", fname)

try:
    inp = open(inp_name)
except Exception:
    print("Could not find file %s" % inp_name)
    sys.exit()

out = open(out_name, "wt")

while 1:
    data = inp.read(BUFFER)
    if not data:
        break
    out.write(data)

inp.close()
out.close()
